//----------------------------------------------------------------------------
// Original file location: C:\cbproj\AudioIO\ASIO\ASIO_dummy.h
// Purpose:
// Once the "Header for an ASIO driver written in plain C"
// (with a non-microsoft compiler, which doesn't support their "thiscall" convention) .
// Now just a small part of DL4YHF's "Audio-I/O-Library", which, besides others,
//     can emulate an ASIO-compatible audio driver (partially, not really conformal) .
// Based on "IExample.h" from Jeff Glatt's nice tutorial "COM in plain C",
//  which used to be at http://www.codeproject.com/KB/COM/com_in_c1.aspx .
// ( whereever you read 'IExample' in this code,
//   consider it being 'the COM object which contains our own ASIO driver') .
//
// Converted into a 'half-way-ASIO-driver-compatible' experimental DLL
//  by Wolfgang Buescher, 2011-10-19 .
//----------------------------------------------------------------------------

#ifndef _ASIO_DRIVER_IN_C_H_
#define _ASIO_DRIVER_IN_C_H_

// ex: #include <initguid.h>  // must be included BEFORE windows.h,
                       // thus it may be TOO LATE to include it here.
                       // Stupid fucking "C"-header-chaos !
                       // "initguid.h" only #defines INITGUID
                       //          and then #includes <guiddef.h> .

// Note: All files in the subdirectory "Steinberg" are subject to
//       Steinberg's ASIO SDK License, and must be copied from THEIR ASIO SDK !
//       ( see details about where to find them in AudioIO_Manual.pdf )
#include "Steinberg/asiosys.h" // must ALWAYS be included BEFORE "asio.h" !
#include "Steinberg/asio.h"    // need the definition of ASIOCallbacks (struct) here

// Note: MY_DLL_NAME, MY_REGISTRY_ASIO_KEY, and MY_ASIO_DRIVER_DESCRIPTION
//       are now defined in the project-specific SWITCHES.H only !

// Our ASIO-COM-Objects's GUID .  MODIFY THE FOLLOWING GUIDS FOR YOUR OWN 'DRIVER' !
// > Every COM object has its own GUID, which is an array of 16 bytes
// > that are different from any other GUID.
// > A COM object's VTable (i.e., interface) also has a GUID.
// Tip: To create a new GUID in BCB, press CTRL-shift-G !
// new: {A5D9CB66-026B-42A8-B085-2AD36A05653F}  [2011-10-25, first experiment]
//  After registering this ASIO driver (similar to "normal" COM-Objects),
//  Search this as a KEY in the registry, under HKEY_LOCAL_MACHINE.SOFTWARE.Classes ,
//   and -for ASIO- also under HKEY_LOCAL_MACHINE.SOFTWARE.ASIO ]
DEFINE_GUID( CLSID_MyAsioDriver, 0xA5D9CB66, 0x026B, 0x42A8, 0xB0, 0x85, 0x2A, 0xD3, 0x6A, 0x05, 0x65, 0x3F );
  // The result of this chaotic macro-trickery can be examined in BCB6 as follows:
  //  In a debug session, WATCH CLSID_MyAsioDriver in HEXADECIMAL form.
  //  If it doesn't look like the GUID defined above, there's something wrong.
  // Furthermore, set a breakpoint in DllRegisterServer(), and watch the "CLSID"
  //  written into the windoze registry (as a string) .
  // Note: There seems to be something fishy about DEFINE_GUID .
  //       Several experts suggested to include initguid.h BEFORE windows.h ,
  //       otherwise DEFINE_GUID may(!) not work as planned.  Holy shit.
// Our COM-VTable's GUID (guess the geeks call it 'IID' = 'Interface-ID')
#if(0)  // only for 'normal' COM servers (ASIO isn't "normal", see below)
DEFINE_GUID( IID_MyAsioDriver,   0xA1ADB1F6, 0x175B, 0x4777, 0x99, 0x45, 0x14, 0x27, 0x97, 0x52, 0x2F, 0x06 );
#else   // Another strange thing about ASIO :
  // Steinberg's ASIO drivers use the same GUID for the 'IID' as for the 'CLSID'.
  // Here's why (copied from Steinberg's own example for an ASIO HOST) :
  //   rc = CoCreateInstance(
  //             lpdrv->clsid,  // [in] REFCLSID rclsid, Class identifier aka special GUID
  //             NULL,          // [in] LPUNKNOWN pUnkOuter,	Object is or isn't part of an aggregate
  //                            // If NULL, indicates that the object is not being created
  //                            // as part of an aggregate (whatever an "aggregate" is..)
  //             CLSCTX_INPROC_SERVER, // [in] DWORD dwClsContext, Context for running executable code
  //             lpdrv->clsid,  // [in] REFIID riid,	Specifies the interface to be used to communicate with the object
  //                            // HERE: SAME AS THE OBJECT'S "CLSID" (why did Steinberg do this ?)
  //             asiodrv);      // LPVOID * ppv, points to where to return the pointer to the requested interface.
# define IID_MyAsioDriver CLSID_MyAsioDriver /* another ASIO-absurdity.. ? ? */
#endif


//---------------------------------------------------------------------------
// ASIO-specific defines (those from Steinberg's "asiosmpl" will be thrown out soon)
#define TESTWAVES 1
// when true, will feed the left input (to host) with
// a sine wave, and the right one with a sawtooth
enum
{
	kAsioSamplesPerBlock = 1024,  // ex: "kAsioSamplesPerBlock" (too imprecise)
           // 1024 samples per block, and 192000 samples per second
           // results in 5.333333 milliseconds to process a single block,
           // which is more reasonable than the ridiculously low blocksize
           // seen in another ASIO driver sample .
	kNumInputs  = 4,
	kNumOutputs = 2
};

//---------------------------------------------------------------------------
// COM-related defines
//
// Define IExample's functions (in a way that works both for our COM DLL
// and any app that #include's this .H file .
// WB: Because this DECLARE_INTERFACE_ - thing was too obfuscating,
//     the following macro-monster was expanded step-by-step
//     to find out what it actually does:
#if(0)  // ORIGINAL code from Jeff Glatt's 'IExample' :
#undef  INTERFACE
#define INTERFACE   TMyAsioDriver     /* ex: IExample */
DECLARE_INTERFACE_ (INTERFACE, IUnknown)
{
	STDMETHOD  (QueryInterface) (THIS_ REFIID, void **) PURE;
	STDMETHOD_ (ULONG, AddRef)	 (THIS) PURE;
	STDMETHOD_ (ULONG, Release) (THIS) PURE;
	// ex (from IExample) : STDMETHOD (SetString) (THIS_ char *) PURE;
	//                      STDMETHOD (GetString) (THIS_ char *, DWORD) PURE;
};
#else  // similar without any fancy obfuscating macro... only "true plain C".
       // The key to expanding DECLARE_INTERFACE_ is in OBJBASE.H,
       // but a cleaner example is in http://www.codeproject.com/KB/COM/com_in_c1.aspx:
#ifndef __thiscall
# define __thiscall /* NOT A KEYWORD, just to mark where Steinberg uses 'thiscall' */
#endif
typedef HRESULT __stdcall QueryInterfacePtr(struct tMyAsioDriver * pDriver, REFIID, void **);
typedef ULONG   __stdcall AddRefPtr(struct tMyAsioDriver * pDriver);
typedef ULONG   __stdcall ReleasePtr(struct tMyAsioDriver * pDriver);
typedef long   __thiscall tAsioVT_init(struct tMyAsioDriver * pDriver, void *sysHandle);
typedef long   __thiscall tAsioVT_getDriverName(struct tMyAsioDriver * pDriver, char *name);
typedef long   __thiscall tAsioVT_getDriverVersion(struct tMyAsioDriver * pDriver);
typedef long   __thiscall tAsioVT_getErrorMessage(struct tMyAsioDriver * pDriver, char *string);
typedef long   __thiscall tAsioVT_start(struct tMyAsioDriver * pDriver);
typedef long   __thiscall tAsioVT_stop(struct tMyAsioDriver * pDriver);
typedef long   __thiscall tAsioVT_getChannels(struct tMyAsioDriver * pDriver, long *numInputChannels, long *numOutputChannels);
typedef long   __thiscall tAsioVT_getLatencies(struct tMyAsioDriver * pDriver, long *inputLatency, long *outputLatency);
typedef long   __thiscall tAsioVT_getBufferSize(struct tMyAsioDriver * pDriver, long *minSize, long *maxSize, long *preferredSize, long *granularity);
typedef long   __thiscall tAsioVT_canSampleRate(struct tMyAsioDriver * pDriver, ASIOSampleRate sampleRate);
typedef long   __thiscall tAsioVT_getSampleRate(struct tMyAsioDriver * pDriver, ASIOSampleRate *sampleRate);
typedef long   __thiscall tAsioVT_setSampleRate(struct tMyAsioDriver * pDriver, ASIOSampleRate sampleRate);
typedef long   __thiscall tAsioVT_getClockSources(struct tMyAsioDriver * pDriver, ASIOClockSource *clocks, long *numSources);
typedef long   __thiscall tAsioVT_setClockSource(struct tMyAsioDriver * pDriver, long iClockSourceIndex );
typedef long   __thiscall tAsioVT_getSamplePosition(struct tMyAsioDriver * pDriver, ASIOSamples *sPos, ASIOTimeStamp *tStamp);
typedef long   __thiscall tAsioVT_getChannelInfo(struct tMyAsioDriver * pDriver, ASIOChannelInfo *info);
typedef long   __thiscall tAsioVT_createBuffers(struct tMyAsioDriver * pDriver, ASIOBufferInfo *bufferInfos, long numChannels, long bufferSize, ASIOCallbacks *callbacks);
typedef long   __thiscall tAsioVT_disposeBuffers(struct tMyAsioDriver * pDriver );
typedef long   __thiscall tAsioVT_controlPanel(  struct tMyAsioDriver * pDriver );
typedef long   __thiscall tAsioVT_future( struct tMyAsioDriver * pDriver, long selector,void *opt);
typedef long   __thiscall tAsioVT_outputReady(struct tMyAsioDriver * pDriver);

typedef struct tMyAsioDriverVtbl{
   QueryInterfacePtr  *QueryInterface;
   AddRefPtr          *AddRef;
   ReleasePtr         *Release;
   // Above: Standard COM-related stuff, using the good old __stdcall convention.
   // Below: Placeholders for ASIO-related stuff, using the ugly 'thiscall' convention.
   // Because 'our' C compiler (Borland or GCC) doesn't support 'thiscall',
   //  it makes no sense to specify a nice complete parameter list here.
   //  See AsioDriverThiscalls_BCB.C for the wrapper functions (in assembler).
   //  Yes indeed, this is REALLY ugly !
   tAsioVT_init  *pAsioThiscall_init;       // ex: virtual ASIOBool init(void *sysHandle);
   tAsioVT_getDriverName  *pAsioThiscall_getDriverName;  // ex: virtual void getDriverName(char *name);
   tAsioVT_getDriverVersion  *pAsioThiscall_getDriverVersion; // ex: virtual long getDriverVersion();
   tAsioVT_getErrorMessage  *pAsioThiscall_getErrorMessage;  // ex: virtual void getErrorMessage(char *string);
   tAsioVT_start  *pAsioThiscall_start;            // ex: virtual ASIOError start();
   tAsioVT_stop   *pAsioThiscall_stop;             // ex: virtual ASIOError stop();
   tAsioVT_getChannels    *pAsioThiscall_getChannels;      // ex: virtual ASIOError getChannels(long *numInputChannels, long *numOutputChannels);
   tAsioVT_getLatencies   *pAsioThiscall_getLatencies;     // ex: virtual ASIOError getLatencies(long *inputLatency, long *outputLatency);
   tAsioVT_getBufferSize  *pAsioThiscall_getBufferSize;    // ex: virtual ASIOError getBufferSize(long *minSize, long *maxSize, long *preferredSize, long *granularity);
   tAsioVT_canSampleRate  *pAsioThiscall_canSampleRate;    // ex: virtual ASIOError canSampleRate(ASIOSampleRate sampleRate);
   tAsioVT_getSampleRate  *pAsioThiscall_getSampleRate;    // ex: virtual ASIOError getSampleRate(ASIOSampleRate *sampleRate);
   tAsioVT_setSampleRate  *pAsioThiscall_setSampleRate;    // ex: virtual ASIOError setSampleRate(ASIOSampleRate sampleRate);
   tAsioVT_getClockSources *pAsioThiscall_getClockSources;  // ex: virtual ASIOError getClockSources(ASIOClockSource *clocks, long *numSources);
   tAsioVT_setClockSource  *pAsioThiscall_setClockSource;   // ex: virtual ASIOError setClockSource(long reference);
   tAsioVT_getSamplePosition  *pAsioThiscall_getSamplePosition; // ex: virtual ASIOError getSamplePosition(ASIOSamples *sPos, ASIOTimeStamp *tStamp);
   tAsioVT_getChannelInfo *pAsioThiscall_getChannelInfo;   // ex: virtual ASIOError getChannelInfo(ASIOChannelInfo *info);
   tAsioVT_createBuffers  *pAsioThiscall_createBuffers;    // ex: virtual ASIOError createBuffers(ASIOBufferInfo *bufferInfos, long numChannels, long bufferSize, ASIOCallbacks *callbacks);
   tAsioVT_disposeBuffers *pAsioThiscall_disposeBuffers;   // ex: virtual ASIOError disposeBuffers();
   tAsioVT_controlPanel   *pAsioThiscall_controlPanel;     // ex: virtual ASIOError controlPanel();
   tAsioVT_future         *pAsioThiscall_future;           // ex: virtual ASIOError future(long selector,void *opt);
   tAsioVT_outputReady    *pAsioThiscall_outputReady;      // ex: virtual ASIOError outputReady();
} IMyAsioDriverVtbl;

typedef struct tMyAsioDriver{
   IMyAsioDriverVtbl *lpVtbl;  // first entry in a COM "object": POINTER to the VTable (not the VTable itself!)

   // our COM-related 'private' properties / members go here:
   DWORD         count;        // reference

   // our 'ASIO-specific' properties / members follow below:
   char sz123ErrorMessage[123+1];
	double dblSamplePosition;
	double dblSampleRate;       // must never be zero because we don't check to avoid divide-by-zero !
	ASIOCallbacks *pCallbacks;
	ASIOTime asioTime;
	ASIOTimeStamp theSystemTime;
	short *inputBuffers[kNumInputs * 2];
	short *outputBuffers[kNumOutputs * 2];
#if TESTWAVES
   double dblSinewavePhase[kNumInputs];
#endif
	long inMap[kNumInputs];
	long outMap[kNumOutputs];
	long nSamplesPerBlock; // ex:  "blockFrames" (too unprecise)  
	long inputLatency;
	long outputLatency;
	long nActiveInputs;
	long nActiveOutputs;
	long toggle;
	BOOL active, started;
	BOOL timeInfoMode, tcRead;

} TMyAsioDriver; // ex: IExample; MyRealIExample;

#endif // use DECLARE_INTERFACE_ (from OBJBASE.H) or not ?

/* Replacement for the ASIO driver's former DllMain (aka "DLL entry point"), */
/*  the following function will be called from the 'real' DllMain :          */
BOOL WINAPI ASIO_DllMain(HINSTANCE instance, DWORD fdwReason, LPVOID lpvReserved);

__declspec(dllexport) HRESULT __stdcall DllRegisterServer(void);
__declspec(dllexport) HRESULT __stdcall DllUnregisterServer(void);

/* Replacements for the 'public' methods if the ASIO driver was a C++ class  */
/* Note that the calling convention of these functions is *NOT* 'thiscall' ! */
void AsioDrv_Construct( TMyAsioDriver *pDriver ); // replaces a constructor (in C++)
void AsioDrv_Destruct(  TMyAsioDriver *pDriver ); // replaces a destructor (in C++)

#endif // _ASIO_DRIVER_IN_C_H_
