// File:    c:\cbproj\ASIO_in_plain_C\AsioDriverThiscalls.h .
// Author:  Wolfgang Buescher (DL4YHF)
// Date:    2011-10-19
// Purpose:
//   Wrapper functions using the 'thiscall' calling convention,
//   used to implement an ASIO driver with a compiler which does NOT
//   support 'thiscall' (for example, Borland C++Builder) .

#ifndef __ASIO_H // include Steinberg's includes if not already included:
# include "Steinberg\asiosys.h"  // must be copied from Steinberg's ASIO SDK !
# include "Steinberg\asio.h"     // must be copied from Steinberg's ASIO SDK !
#endif

#ifndef __thiscall
# define __thiscall /* NOT A KEYWORD, just to mark where Steinberg uses 'thiscall' */
#endif


// Functions implemented 'in plain C' somewhere in 'my' ASIO driver.
// Much in contrast to Steinberg's ASIO, they use a calling convention
// which is supported by ANY C-compiler:  __stdcall !
// We use the prefix 'AsioDrv_' here, because there may be other functions
// like 'start()' or 'stop()' in our project which don't have anything to do
// with (Steinberg's) ASIO .
// The first function argument would be the 'this'-pointer, if we could use C++
//   (which is impossible because our C++ compiler uses a different calling
//   convention, which is NOT compatible with Microsoft Visual C++).
//   'IMyAsioDriver' which mimicks a COM-object, see ASIO_driver_in_c.h .
ASIOBool  __stdcall AsioDrv_init(struct tMyAsioDriver * pThis, void *sysHandle);
void      __stdcall AsioDrv_getDriverName(struct tMyAsioDriver * pThis,
                      char *name);   // [out] max 32 bytes incl. terminating zero
long      __stdcall AsioDrv_getDriverVersion(struct tMyAsioDriver * pThis);
void      __stdcall AsioDrv_getErrorMessage(struct tMyAsioDriver * pThis,
                      char *string); // [out] max 124 bytes incl. trailing zero
ASIOError __stdcall AsioDrv_start(struct tMyAsioDriver * pThis);
ASIOError __stdcall AsioDrv_stop(struct tMyAsioDriver * pThis);
ASIOError __stdcall AsioDrv_getChannels(struct tMyAsioDriver * pThis, long *numInputChannels, long *numOutputChannels);
ASIOError __stdcall AsioDrv_getLatencies(struct tMyAsioDriver * pThis, long *inputLatency, long *outputLatency);
ASIOError __stdcall AsioDrv_getBufferSize(struct tMyAsioDriver * pThis, long *minSize, long *maxSize, long *preferredSize, long *granularity);
ASIOError __stdcall AsioDrv_canSampleRate(struct tMyAsioDriver * pThis, ASIOSampleRate sampleRate);
ASIOError __stdcall AsioDrv_getSampleRate(struct tMyAsioDriver * pThis, ASIOSampleRate *sampleRate);
ASIOError __stdcall AsioDrv_setSampleRate(struct tMyAsioDriver * pThis, ASIOSampleRate sampleRate);
ASIOError __stdcall AsioDrv_getClockSources(struct tMyAsioDriver * pThis, ASIOClockSource *clocks, long *numSources);
ASIOError __stdcall AsioDrv_setClockSource(struct tMyAsioDriver * pThis, long reference);
ASIOError __stdcall AsioDrv_getSamplePosition(struct tMyAsioDriver * pThis, ASIOSamples *sPos, ASIOTimeStamp *tStamp);
ASIOError __stdcall AsioDrv_getChannelInfo(struct tMyAsioDriver * pThis, ASIOChannelInfo *info);
ASIOError __stdcall AsioDrv_createBuffers(struct tMyAsioDriver * pThis, ASIOBufferInfo *bufferInfos, long numChannels, long bufferSize, ASIOCallbacks *callbacks);
ASIOError __stdcall AsioDrv_disposeBuffers(struct tMyAsioDriver * pThis);
ASIOError __stdcall AsioDrv_controlPanel(struct tMyAsioDriver * pThis);
ASIOError __stdcall AsioDrv_future(struct tMyAsioDriver * pThis, long selector,void *opt);
ASIOError __stdcall AsioDrv_outputReady(struct tMyAsioDriver * pThis);



// Functions implemented somewhere in ASSEMBLY LANGUAGE with the 'thiscall'
// calling convention. Their only purpose is to call 'our ASIO driver',
// using the __stdcall calling convention.
// Note: It makes only little sense to specify an argument list for these functions here,
//       because the C compiler would then generate its own 'framework'
//       which would definitely NOT be compatible with microsoft's 'thiscall'.
// The return value is always in EAX, regardless of the type
//    (ASIOBool, void, long, or ASIOError) .
// Here: ORDERED BY THEIR APPEARANCE IN AN ASIO DRIVER'S VTABLE .
long AsioThiscall_init(struct tMyAsioDriver * pThis, void *sysHandle);
long AsioThiscall_getDriverName(struct tMyAsioDriver * pThis, char *name);
long AsioThiscall_getDriverVersion(struct tMyAsioDriver * pThis);
long AsioThiscall_getErrorMessage(struct tMyAsioDriver * pThis, char *string);
long AsioThiscall_start(struct tMyAsioDriver * pThis);
long AsioThiscall_stop(struct tMyAsioDriver * pThis);
long AsioThiscall_getChannels(struct tMyAsioDriver * pThis, long *numInputChannels, long *numOutputChannels);
long AsioThiscall_getLatencies(struct tMyAsioDriver * pThis, long *inputLatency, long *outputLatency);
long AsioThiscall_getBufferSize(struct tMyAsioDriver * pThis, long *minSize, long *maxSize, long *preferredSize, long *granularity);
long AsioThiscall_canSampleRate(struct tMyAsioDriver * pThis, ASIOSampleRate sampleRate);
long AsioThiscall_getSampleRate(struct tMyAsioDriver * pThis, ASIOSampleRate *sampleRate);
long AsioThiscall_setSampleRate(struct tMyAsioDriver * pThis, ASIOSampleRate sampleRate);
long AsioThiscall_getClockSources(struct tMyAsioDriver * pThis, ASIOClockSource *clocks, long *numSources);
long AsioThiscall_setClockSource(struct tMyAsioDriver * pThis, long reference);
long AsioThiscall_getSamplePosition(struct tMyAsioDriver * pThis, ASIOSamples *sPos, ASIOTimeStamp *tStamp);
long AsioThiscall_getChannelInfo(struct tMyAsioDriver * pThis, ASIOChannelInfo *info);
long AsioThiscall_createBuffers(struct tMyAsioDriver * pThis, ASIOBufferInfo *bufferInfos, long numChannels, long bufferSize, ASIOCallbacks *callbacks);
long AsioThiscall_disposeBuffers(struct tMyAsioDriver * pThis);
long AsioThiscall_controlPanel(struct tMyAsioDriver * pThis  );
long AsioThiscall_future(struct tMyAsioDriver * pThis, long selector,void *opt);
long AsioThiscall_outputReady(struct tMyAsioDriver * pThis);
