//---------------------------------------------------------------------------
// File:    c:\cbproj\AudioIO\ASIO\AsioDriverThiscalls_BCB.c
// Author:  Wolfgang Buescher (DL4YHF)
// Date:    2011-10-19
// Purpose:
//    Wrapper functions using the 'thiscall' calling convention,
//    used to implement an ASIO driver with a compiler which does NOT
//    support 'thiscall'.
//    Original file  : c:\cbproj\ASIO_in_plain_C\AsioDriverThiscalls_BCB.c
//    Modified copies: c:\cbproj\AudioIO\ASIO\AsioDriverThiscalls_BCB.c
// Here:  Borland C++ Builder V6 (Enterprise, because the standard edition
//        doesn't allow inline assembler ... what a shame..)
//        In comparison with Borland's inline assembly,
//        the GNU Compiler's inline assembly looks really ugly
//        so if you seriously plan to get a similar module working with GCC:
//        GOOD LUCK !
//
//---------------------------------------------------------------------------

#include "AsioDriverThiscalls.h"

/*
  THISCALL DEFINITION (from Ross Bencina's IASIOThiscallResolver.cpp)

    A number of definitions of the thiscall calling convention are floating
    around the internet. The following definition has been validated against
    output from the MSVC++ compiler:
    For non-vararg functions, thiscall works as follows: the object (this)
    pointer is passed in ECX. All arguments are passed on the stack in
    right to left order. The return value is placed in EAX. The callee
    clears the passed arguments from the stack.

  THISCALL definition (from Microsoft)
    This is the default calling convention used by C++ member functions
    that do not use variable arguments.
    Under thiscall, the callee cleans the stack, which is impossible for
    vararg functions. Arguments are pushed on the stack from right to left,
    with the this pointer being passed via register ECX on the x86 architecture.
    The thiscall calling convention cannot be explicitly specified in a program,
    because thiscall is not a keyword.
    vararg member functions use the __cdecl calling convention. All function
    arguments are pushed on the stack, with the this pointer placed on the stack last
    Because this calling convention applies only to C++, there is no C name decoration scheme.

  stdcall (from Wikipedia)
    The stdcall[2] calling convention is a variation on the pascal calling convention
    in which the callee is responsible for cleaning up the stack,
    but the parameters are pushed onto the stack in right-to-left order,
    as in the _cdecl calling convention.
    Registers EAX, ECX, and EDX are designated for use within the function.
    Return values are stored in the EAX register.
*/




//---------------------------------------------------------------------------
// Functions using the  'thiscall' calling convention with NO parameters
//      on the stack (only the 'this'-pointer in the ECX register)
//---------------------------------------------------------------------------

#pragma warn -8057 // "parameter pThis is never used" - must be ignored

long AsioThiscall_getDriverVersion(struct tMyAsioDriver * pThis)
{ // The compiler( Borland C++Builder) inserts the following boilerplate code here,
  //     so we must pop these registers from the stack before returning:
  // push ebp       // boilerplate code for the 'function entry', from C compiler
  // mov  ebp,esp   // boilerplate code for the 'function entry', from C compiler
  // push ebx       // boilerplate code for the 'function entry', from C compiler (BCB6)
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { // At THIS point, after the above boilerplate code was executed:
     // ("oldest" stack item at the end of the list, with the LARGEST stack address,
     //  because the stack pointer -ESP- is DECREMENTED when pushing new items)
     //    ECX     = 0x0012FF34  ('this'-pointer, formally the FIRST (invisible) function argument)
     //    EBX     = 0x7FFD9000  (must be important, or WHY does BCB emit code to save it ?)
     //    ESP     = 0x0012FF20  (stack pointer, when tested)
     //    EBP     = 0x0012FF24  (base pointer, when tested)
     //    EBP[-4] = ESP[ 0] = original value of EBX    (value 0x7FFD9000 @ 0012FF20 when tested)
     //    EBP[ 0] = ESP[+4] = original value of EBP    (value 0x0012FF8C @ 0012FF24 when tested)
     //    EBP[+4] = ESP[+8] = caller's return address  (value 0x0040118D @ 0012FF28 when tested)
     // Make a stdcall with ONE argument on the stack,
     //                here:  only the "this"-pointer .
     push ecx   // push the 'this'-pointer (in ECX from Microsoft's "thiscall")
     call AsioDrv_getDriverVersion  // __stdcall, pops ONE 32-bit value from stack, result in EAX
     // At this point, stdcall AsioDrv_getDriverVersion has cleaned up its function argument.
     // In other words, it has 'undone' the push ecx above, but nothing else.
     pop ebx    // restore EBX (counterpart for the compiler-generated function entry code mentioned above)
     pop ebp    // restore EBP (also part of the function-entry-code)
     ret        // return WITHOUT popping anything (the argument 'pThis' is NOT on the stack but in ECX)
     // Return value still in EAX at this point
   }
  // After the above asm-block, the compiler added the following code
  //  (which we don't want, since we already returned, it doesn't really matter)
  // pop ebx
  // pop ebp
  // ret
  return _EAX; // make the compiler happy ("function should return a value")
} // end AsioThiscall_getDriverVersion()


long AsioThiscall_start(struct tMyAsioDriver * pThis)
{ // For an explanation, see AsioThiscall_getDriverVersion() [similar arguments]
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { push ecx  // "this"-pointer passed via ECX in Microsoft's 'thiscall'
     call AsioDrv_start
     pop ebx
     pop ebp
     ret
   }
  return _EAX; // make the compiler happy ("function should return a value")
}
long AsioThiscall_stop(struct tMyAsioDriver * pThis)
{ // For an explanation, see AsioThiscall_getDriverVersion() [similar arguments]
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { push ecx
     call AsioDrv_stop
     pop ebx
     pop ebp
     ret
   }
  return _EAX; // make the compiler happy ("function should return a value")
}
long AsioThiscall_disposeBuffers(struct tMyAsioDriver * pThis)
{ // For an explanation, see AsioThiscall_getDriverVersion() [similar arguments]
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { push ecx  // "this"-pointer passed via ECX in Microsoft's 'thiscall'
     call AsioDrv_disposeBuffers
     pop ebx
     pop ebp
     ret
   }
  return _EAX; // make the compiler happy ("function should return a value")
}
long AsioThiscall_controlPanel(struct tMyAsioDriver * pThis)
{ // For an explanation, see AsioThiscall_getDriverVersion() [similar arguments]
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { push ecx  // "this"-pointer passed via ECX in Microsoft's 'thiscall'
     call AsioDrv_controlPanel
     pop ebx
     pop ebp
     ret
   }
  return _EAX; // make the compiler happy ("function should return a value")
}
long AsioThiscall_outputReady(struct tMyAsioDriver * pThis)
{ // For an explanation, see AsioThiscall_getDriverVersion() [similar arguments]
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { push ecx  // "this"-pointer passed via ECX in Microsoft's 'thiscall'
     call AsioDrv_outputReady
     pop ebx
     pop ebp
     ret
   }
  return _EAX; // make the compiler happy ("function should return a value")
}




//---------------------------------------------------------------------------
// Functions using the 'thiscall' calling convention with ONE 32-bit parameter
//           on the stack, plus the 'this'-pointer in the ECX register .
//---------------------------------------------------------------------------

long AsioThiscall_init(struct tMyAsioDriver * pThis, void *sysHandle)
{
  // The compiler( Borland C++Builder) inserts the following boilerplate code here,
  //     so we must pop these registers from the stack before returning:
  // push ebp       // boilerplate code for the 'function entry', from C compiler
  // mov  ebp,esp   // boilerplate code for the 'function entry', from C compiler
  // push ebx       // boilerplate code for the 'function entry', from C compiler (BCB6)
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { // At THIS point, after the above boilerplate code was executed:
     // ("oldest" stack item at the end of the list, with the LARGEST stack address,
     //  because the stack pointer -ESP- is DECREMENTED when pushing new items)
     //    ECX     = 0x0012FF34  ('this'-pointer, formally the FIRST (invisible) function argument)
     //    EBX     = 0x7FFD9000  (must be important, or WHY does BCB emit code to save it ?)
     //    ESP     = 0x0012FF20  (stack pointer, when tested)
     //    EBP     = 0x0012FF24  (base pointer, when tested)
     //    EBP[-4] = ESP[ 0] = original value of EBX    (value 0x7FFD9000 @ 0012FF20 when tested)
     //    EBP[ 0] = ESP[+4] = original value of EBP    (value 0x0012FF8C @ 0012FF24 when tested)
     //    EBP[+4] = ESP[+8] = caller's return address  (value 0x0040118D @ 0012FF28 when tested)
     //    EBP[+8] = ESP[+12]= rightmost argument first, pushed FIRST by the thiscall-caller,
     //                       here: void *sysHandle     (value 0x00000000 @ 0012FF2C when tested)
     // Make a stdcall with TWO arguments on the stack,
     //   parameters are pushed onto the stack in right-to-left order, i.e.
     //   RIGHTMOST argument pushed first,
     //   LEFTMOST argument  pushed last (here: the "this"-pointer) .
     push dword ptr [ebp+8] // push the rightmost function argument (here: sysHandle)
     push ecx           // push the 'this'-pointer (in ECX from 'thiscall')
     call AsioDrv_init  // __stdcall, pops TWO 32-bit values from stack, result in EAX
     // stdcall AsioDrv_init has cleaned up its TWO function arguments.
     // Here, in the wrapper which emulates a C++ member function with 'thiscall',
     // the callee(!) must clean up one argument less, because the 'this'-ptr
     // was passed in ECX, not on the stack !
     pop ebx    // restore EBX (counterpart for the compiler-generated function entry code mentioned above)
     pop ebp    // restore EBP (also part of the function-entry-code)
     ret 4      // return and discard 4 bytes (=ONLY 2nd to last function argument,
                // NOT the first argument because in 'thiscall', pThis is passed in ECX)
     // Return value still in EAX at this point
   }
  // After the above asm-block, the compiler added the following code
  //  (which we don't want, since we already returned, it doesn't really matter)
  // pop ebx
  // pop ebp
  // ret

  return _EAX; // just to make the compiler happy ("function should return a value")
} // end AsioThiscall_init()

long AsioThiscall_getDriverName(struct tMyAsioDriver * pThis, char *name)
{ // Similar wrapper as all other ASIO member functions ("methods" in C++)
  //  with ONE argument in the visible argument list (plus one for the THIS-pointer) .
  // For details and an explanation, see AsioThiscall_init() .
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { push dword ptr [ebp+8] // push the rightmost function argument (here: char* name)
     push ecx               // push the 'this'-pointer (in ECX from 'thiscall')
     call AsioDrv_getDriverName // __stdcall, pops TWO 32-bit values from stack, result in EAX
     pop ebx    // restore EBX (counterpart for the compiler-generated function entry code)
     pop ebp    // restore EBP (also part of the function-entry-code)
     ret 4      // return and discard 4 bytes (the 1st argument, 'pThis', is NOT on the stack but in ECX)
   }
  return _EAX; // just to make the compiler happy ("function should return a value")
} // end AsioThiscall_getDriverName()

long AsioThiscall_getErrorMessage(struct tMyAsioDriver * pThis, char *string)
{ // Similar wrapper as all other ASIO member functions ("methods" in C++)
  //  with ONE argument in the visible argument list (plus one for the THIS-pointer) .
  // For details and an explanation, see AsioThiscall_init() .
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { push dword ptr [ebp+8] // push the rightmost function argument (here: char* string)
     push ecx               // push the 'this'-pointer (in ECX from 'thiscall')
     call AsioDrv_getErrorMessage // __stdcall, pops TWO 32-bit values from stack, result in EAX
     pop ebx    // restore EBX (counterpart for the compiler-generated function entry code)
     pop ebp    // restore EBP (also part of the function-entry-code)
     ret 4      // return and discard 4 bytes (the 1st argument, 'pThis', is NOT on the stack but in ECX)
   }
  return _EAX; // just to make the compiler happy ("function should return a value")
} // end AsioThiscall_getErrorMessage()

long AsioThiscall_setClockSource(struct tMyAsioDriver * pThis, long reference)
{ // For details and an explanation, see AsioThiscall_init() .
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { push dword ptr [ebp+8] // push the rightmost function argument (here: long reference)
     push ecx               // push the 'this'-pointer (in ECX from 'thiscall')
     call AsioDrv_getErrorMessage // __stdcall, pops TWO 32-bit values from stack, result in EAX
     pop ebx    // restore EBX (counterpart for the compiler-generated function entry code)
     pop ebp    // restore EBP (also part of the function-entry-code)
     ret 4      // return and discard 4 bytes (the 1st argument, 'pThis', is NOT on the stack but in ECX)
   }
  return _EAX; // just to make the compiler happy ("function should return a value")
} // end AsioThiscall_setClockSource()

long AsioThiscall_getChannelInfo(struct tMyAsioDriver * pThis, ASIOChannelInfo *info)
{ // For details and an explanation, see AsioThiscall_init() .
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { push dword ptr [ebp+8] // push the rightmost function argument (here: *info)
     push ecx               // push the 'this'-pointer (in ECX from 'thiscall')
     call AsioDrv_getChannelInfo // __stdcall, pops TWO 32-bit values from stack, result in EAX
     pop ebx    // restore EBX (counterpart for the compiler-generated function entry code)
     pop ebp    // restore EBP (also part of the function-entry-code)
     ret 4      // return and discard 4 bytes (the 1st argument, 'pThis', is NOT on the stack but in ECX)
   }
  return _EAX; // just to make the compiler happy ("function should return a value")
} // end AsioThiscall_getChannelInfo()

long AsioThiscall_getSampleRate(struct tMyAsioDriver * pThis, ASIOSampleRate *sampleRate)
{ // Similar wrapper as all other ASIO member functions ("methods" in C++)
  //  with ONE argument in the visible argument list (plus one for the THIS-pointer) .
  // The function argument is NOT a 64-bit floating number here,
  //  but a 32-bit POINTER to a double. Thus it does NOT fall
  //  into the same category as AsioThiscall_canSampleRate()
  //  and AsioThiscall_setSampleRate() !
  // For details and an explanation, see AsioThiscall_init() .
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { push dword ptr [ebp+8] // push the rightmost function argument (here: a POINTER, i.e. 32-bit address)
     push ecx               // push the 'this'-pointer (in ECX from 'thiscall')
     call AsioDrv_getSampleRate // __stdcall, pops TWO 32-bit values from stack, result in EAX
     pop ebx    // restore EBX (counterpart for the compiler-generated function entry code)
     pop ebp    // restore EBP (also part of the function-entry-code)
     ret 4      // return and discard 4 bytes (the 1st argument, 'pThis', is NOT on the stack but in ECX)
   }
  return _EAX; // just to make the compiler happy ("function should return a value")
} // end AsioThiscall_getSampleRate()





//---------------------------------------------------------------------------
// Functions using the 'thiscall' calling convention with TWO 32-bit parameters
//           on the stack, plus the 'this'-pointer in the ECX register .
//---------------------------------------------------------------------------

long AsioThiscall_getChannels(struct tMyAsioDriver * pThis, long *numInputChannels, long *numOutputChannels)
{
  // The compiler( Borland C++Builder) inserts the following boilerplate code here,
  //     so we must pop these registers from the stack before returning:
  // push ebp       // boilerplate code for the 'function entry', from C compiler
  // mov  ebp,esp   // boilerplate code for the 'function entry', from C compiler
  // push ebx       // boilerplate code for the 'function entry', from C compiler (BCB6)
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { // At THIS point, after the above boilerplate code was executed:
     // ("oldest" stack item at the end of the list, with the LARGEST stack address,
     //  because the stack pointer -ESP- is PRE-DECREMENTED when pushing an item)
     //    ECX     = 0x0012FF34  ('this'-pointer, formally the FIRST (invisible) function argument)
     //    EBX     = 0x7FFD9000  (must be important, or WHY does BCB emit code to save it ?)
     //    ESP     = 0x0012FF20  (stack pointer, when tested)
     //    EBP     = 0x0012FF24  (base pointer, when tested)
     //    EBP[-4] = ESP[ 0] = original value of EBX
     //    EBP[ 0] = ESP[+4] = original value of EBP
     //    EBP[+4] = ESP[+8] = caller's return address
     //    EBP[+8] = ESP[+12]= leftmost argument, pushed SECOND by the thiscall-caller,
     //                       here: numInputChannels
     //    EBP[+12]= ESP[+16]= rightmost argument, pushed FIRST by the thiscall-caller,
     //                       here: numOutputChannels
     // Make a stdcall with THREE arguments on the stack,
     //   parameters are pushed onto the stack in right-to-left order, i.e.
     //   RIGHTMOST argument pushed first (here: numOutputChannels),
     //    followed numInputChannels,
     //   LEFTMOST argument  pushed last (here: the "this"-pointer) .
     push dword ptr [ebp+12] // push the rightmost function argument (here: numOutputChannels)
     push dword ptr [ebp+8]  // push next-left argument (here: numInputChannels)
     push ecx                // push the 'this'-pointer (in ECX from 'thiscall')
     call AsioDrv_getChannels // __stdcall, pops THREE 32-bit values from stack, result in EAX
     // Arrived here, stdcall AsioDrv_getChannels has cleaned up its THREE function arguments.
     // Here, in the wrapper which emulates a C++ member function with 'thiscall',
     // the callee(!) must clean up one argument less, because the 'this'-ptr
     // was passed in ECX, not on the stack !
     pop ebx    // restore EBX (counterpart for the compiler-generated function entry code mentioned above)
     pop ebp    // restore EBP (also part of the function-entry-code)
     ret 8      // return and discard 8 bytes (=ONLY the 2nd to last function argument,
                // NOT the first argument because in 'thiscall', pThis is passed in ECX)
     // Return value still in EAX at this point
   }
  // After the above asm-block, the compiler added the following code
  //  (which we don't want, since we already returned, it doesn't really matter)
  // pop ebx
  // pop ebp
  // ret

  return _EAX; // make the compiler happy ("function should return a value")
} // end AsioThiscall_getChannels()

long AsioThiscall_getLatencies(struct tMyAsioDriver * pThis, long *inputLatency, long *outputLatency)
{ // For an explanation, see AsioThiscall_getChannels() [similar arguments]
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { push dword ptr [ebp+12] // push the rightmost function argument (here: long* outputLatency)
     push dword ptr [ebp+8]  // push next-left argument (here: long* inputLatency)
     push ecx                // push the 'this'-pointer (in ECX from 'thiscall')
     call AsioDrv_getLatencies
     pop ebx
     pop ebp
     ret 8
   }
  return _EAX; // make the compiler happy ("function should return a value")
} // end AsioThiscall_getLatencies()

long AsioThiscall_getClockSources(struct tMyAsioDriver * pThis, ASIOClockSource *clocks, long *numSources)
{ // For an explanation, see AsioThiscall_getChannels() [similar arguments]
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { push dword ptr [ebp+12] // push the rightmost function argument
     push dword ptr [ebp+8]  // push next-left argument
     push ecx                // push the 'this'-pointer (in ECX from 'thiscall')
     call AsioDrv_getClockSources
     pop ebx
     pop ebp
     ret 8
   }
  return _EAX; // make the compiler happy ("function should return a value")
} // end AsioThiscall_getClockSources()

long AsioThiscall_getSamplePosition(struct tMyAsioDriver * pThis, ASIOSamples *sPos, ASIOTimeStamp *tStamp)
{ // For an explanation, see AsioThiscall_getChannels() [similar arguments]
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { push dword ptr [ebp+12] // push the rightmost function argument
     push dword ptr [ebp+8]  // push next-left argument
     push ecx                // push the 'this'-pointer (in ECX from 'thiscall')
     call AsioDrv_getSamplePosition
     pop ebx
     pop ebp
     ret 8
   }
  return _EAX; // make the compiler happy ("function should return a value")
} // end AsioThiscall_getSamplePosition()

long AsioThiscall_future(struct tMyAsioDriver * pThis, long selector,void *opt)
{ // For an explanation, see AsioThiscall_getChannels() [similar arguments]
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { push dword ptr [ebp+12] // push the rightmost function argument
     push dword ptr [ebp+8]  // push next-left argument
     push ecx                // push the 'this'-pointer (in ECX from 'thiscall')
     call AsioDrv_future
     pop ebx
     pop ebp
     ret 8
   }
  return _EAX; // make the compiler happy ("function should return a value")
} // end AsioThiscall_future(long selector,void *opt)

//---------------------------------------------------------------------------
// Functions using the 'thiscall' calling convention with FOUR 32-bit parameters
//           on the stack, plus the 'this'-pointer in the ECX register .
//---------------------------------------------------------------------------

long AsioThiscall_getBufferSize(struct tMyAsioDriver * pThis,long *minSize,long *maxSize,long *preferredSize,long *granularity)
{ // The compiler( Borland C++Builder) inserts the following boilerplate code here,
  //     so we must pop these registers from the stack before returning:
  // push ebp       // boilerplate code for the 'function entry', from C compiler
  // mov  ebp,esp   // boilerplate code for the 'function entry', from C compiler
  // push ebx       // boilerplate code for the 'function entry', from C compiler (BCB6)
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { // At THIS point, after the above boilerplate code was executed:
     // ("oldest" stack item at the end of the list, with the LARGEST stack address,
     //  because the stack pointer -ESP- is PRE-DECREMENTED when pushing an item)
     //    ECX     = 0x0012FF34  ('this'-pointer, formally the FIRST (invisible) function argument)
     //    EBX     = 0x7FFD9000  (must be important, or WHY does BCB emit code to save it ?)
     //    ESP     = 0x0012FF20  (stack pointer, when tested)
     //    EBP     = 0x0012FF24  (base pointer, when tested)
     //    EBP[-4] = ESP[ 0] = original value of EBX
     //    EBP[ 0] = ESP[+4] = original value of EBP
     //    EBP[+4] = ESP[+8] = caller's return address
     //    EBP[+8] = ESP[+12]= leftmost argument, pushed FOURTH by the thiscall-caller,
     //                       here: minSize
     //    EBP[+12]= ESP[+16]= 2nd-left argument, pushed THIRD by the thiscall-caller,
     //                       here: maxSize
     //    EBP[+16]= ESP[+20]= 3rd-left argument, pushed SECOND by the thiscall-caller,
     //                       here: preferredSize
     //    EBP[+20]= ESP[+24]= rightmost argument, pushed FIRST by the thiscall-caller,
     //                       here: granularity
     // Make a stdcall with FIVE 32-bit arguments on the stack,
     //   parameters are pushed onto the stack in right-to-left order, i.e.
     //   RIGHTMOST argument pushed first (here: numOutputChannels),
     //   LEFTMOST argument  pushed last (here: the "this"-pointer) .
     push dword ptr [ebp+20] // push the rightmost function argument (here: granularity)
     push dword ptr [ebp+16] // push next argument (here: preferredSize)
     push dword ptr [ebp+12] // push next argument (here: maxSize)
     push dword ptr [ebp+8]  // push next argument (here: minSize)
     push ecx                // push the 'this'-pointer (in ECX from 'thiscall')
     call AsioDrv_getBufferSize // __stdcall, pops FIVE 32-bit values from stack, result in EAX
     // Arrived here, stdcall AsioDrv_getBufferSize has cleaned up its FIVE function arguments.
     // Here, in the wrapper which emulates a C++ member function with 'thiscall',
     // the callee(!) must clean up one argument less, because the 'this'-ptr
     // was passed in ECX, not on the stack !
     pop ebx    // restore EBX (counterpart for the compiler-generated function entry code mentioned above)
     pop ebp    // restore EBP (also part of the function-entry-code)
     ret 16     // return and discard 16 bytes (=ONLY the 2nd to last function argument,
                // NOT the first argument because in 'thiscall', pThis is passed in ECX)
     // Return value still in EAX at this point
   }
  // After the above asm-block, the compiler added the following code
  //  (which we don't want, since we already returned, it doesn't really matter)
  // pop ebx
  // pop ebp
  // ret
  return _EAX; // make the compiler happy ("function should return a value")
} // end AsioThiscall_getBufferSize()

long AsioThiscall_createBuffers(struct tMyAsioDriver * pThis,ASIOBufferInfo *bufferInfos, long numChannels, long bufferSize, ASIOCallbacks *callbacks)
{ // For an explanation, see AsioThiscall_getBufferSize() [similar arguments]
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { push dword ptr [ebp+20] // push the rightmost function argument FIRST....
     push dword ptr [ebp+16]
     push dword ptr [ebp+12]
     push dword ptr [ebp+8]
     push ecx                // push the 'this'-pointer (in ECX from 'thiscall')
     call AsioDrv_createBuffers
     pop ebx
     pop ebp
     ret 16
   }
  return _EAX; // make the compiler happy ("function should return a value")
} // end AsioThiscall_createBuffers()


//---------------------------------------------------------------------------
// Functions using the 'thiscall' calling convention with ONE 64-bit parameter
//      on the stack ("double"), plus the 'this'-pointer in the ECX register .
// Info: When pushing a 64-bit 'double' value on the stack,
//       Borland C++ Builder produced the following code (for Win32) :
//   push  dword ptr [ebp-0x04]  ; most significant 32 bits of 'double' value pushed FIRST
//                               ; so it will be at the HIGHER ADDRESS on the stack
//   push  dword ptr [ebp-0x08]  ; least significant 32 bits of 'double' value pushed SECOND
//                               ; so it will be at the LOWER ADDRESS on the stack
//   In memory, a 64-bit 'double' value begins (at the lowest address, at least DWORD-aligned)
//   with the least significant BYTE of the mantissa,
//   and ends (at the highest address) with the most significant byte of the EXPONENT.
//---------------------------------------------------------------------------

long AsioThiscall_canSampleRate(struct tMyAsioDriver * pThis, ASIOSampleRate sampleRate)
{
  // The compiler( Borland C++Builder) inserts the following boilerplate code here,
  //     so we must pop these registers from the stack before returning:
  // push ebp       // boilerplate code for the 'function entry', from C compiler
  // mov  ebp,esp   // boilerplate code for the 'function entry', from C compiler
  // push ebx       // boilerplate code for the 'function entry', from C compiler (BCB6)
  asm // Note the boilerplate code inserted by the C-compiler BEFORE this asm !
   { // At THIS point, after the above boilerplate code was executed:
     // ("oldest" stack item at the end of the list, with the LARGEST stack address,
     //  because the stack pointer -ESP- is PRE-DECREMENTED when pushing an item)
     //    ECX     = 0x0012FF34  ('this'-pointer, formally the FIRST (invisible) function argument)
     //    EBX     = 0x7FFD9000  (must be important, or WHY does BCB emit code to save it ?)
     //    ESP     = 0x0012FF20  (stack pointer, when tested)
     //    EBP     = 0x0012FF24  (base pointer, when tested)
     //    EBP[-4] = ESP[ 0] = original value of EBX
     //    EBP[ 0] = ESP[+4] = original value of EBP
     //    EBP[+4] = ESP[+8] = caller's return address
     //    EBP[+8] = ESP[+12]= least significant 32 bits of 'double' value (only MANTISSA),
     //                        hopefully pushed SECOND by the thiscall-caller,
     //    EBP[+12]= ESP[+16]= most significant 32 bits of 'double' value (incl. EXPONENT),
     //                        hopefully pushed FIRST by the thiscall-caller.
     // Make a stdcall with one 64-bit, and one 32-bit argument on the stack.
     //   Parameters are pushed onto the stack in right-to-left order, i.e.
     //   RIGHTMOST argument pushed first (here: sampleRate, MSW pushed first, LSW pushed 2nd),
     //   LEFTMOST  argument pushed last (here: the "this"-pointer) .
     push dword ptr [ebp+12] // push rightmost function argument,  here:  MOST SIGNIFICANT BITS of the 'double' value
     push dword ptr [ebp+8 ] // 2nd part of the rightmost argument, here: LEAST SIGNIFICANT BITS of the 'double' value
     push ecx           // push the 'this'-pointer (in ECX from 'thiscall')
     call AsioDrv_canSampleRate // __stdcall, pops one 32-bit address ("this") and one 64-bit value off the stack,
                                // and places the result in EAX .
     // Arrived here, stdcall AsioDrv_canSampleRate has removed ITS function arguments from the stack.
     // Here, in the wrapper which emulates a C++ member function with 'thiscall',
     // the callee(!) must clean up one argument less, because the 'this'-ptr
     // was passed in ECX, not on the stack !
     pop ebx    // restore EBX (counterpart for the compiler-generated function entry code mentioned above)
     pop ebp    // restore EBP (also part of the function-entry-code)
     ret 8      // return and discard 8 bytes (=ONLY the 2nd to last function argument,
                // NOT the first argument because in 'thiscall', pThis is passed in ECX)
     // Return value still in EAX at this point
   }
  // After the above asm-block, the compiler added the following code
  //  (which we don't want, since we already returned, it doesn't really matter)
  // pop ebx
  // pop ebp
  // ret
  return _EAX; // make the compiler happy ("function should return a value")
} // end AsioThiscall_canSampleRate()

long AsioThiscall_setSampleRate(struct tMyAsioDriver * pThis, ASIOSampleRate sampleRate)
{ // For an explanation, see AsioThiscall_canSampleRate() [similar arguments]
  asm
   { push dword ptr [ebp+12]
     push dword ptr [ebp+8 ]
     push ecx
     call AsioDrv_setSampleRate
     pop ebx
     pop ebp
     ret 8
   }
  return _EAX; // make the compiler happy ("function should return a value")
} // end AsioThiscall_setSampleRate()




/* EOF <AsioDriverThiscalls_BCB.c> */